; LittleLights to serial converter
; version 1.0 date 2006-01-14
; Copyright 2005-2006 Stefan Schuermans <1stein@schuermans.info>
; Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
; a BlinkenArea project - http://www.blinkenarea.org/

; clock frequency: 12 MHz (external crystal)

; PA0..1: crystal
; PA2: reset
; PB0..7: output to 74HC4094s
; PD0: serial input (RX)
; PD1: serial output (TX)
; PD2: output to 9th chain of 74HC4094s
; PD3..4: unused
; PD5: strobe for 74HC4094s
; PD6: clock for 74HC4094s

; serial protocol: 115200, 8N1
;   input:
;     message-format: MCUF 18x8-1/256
;     message: <header> <data>
;     header: <magic> <height> <width> <channels> <maxval>
;     magic: 0x23 0x54 0x26 0x66
;     height: 0x00 0x08
;     width: 0x00 0x12
;     channels: 0x00 0x01
;     maxval: 0x00 0xFF
;     data: <line0> ... <line7>
;     line: <pixel0> ... <pixel17>
;     pixel: 0x00..0xFF



.INCLUDE        "tn2313def.inc"



; IO pins
.equ    PORT_CLK        =       PORTD           ; clock: PD6
.equ    BIT_CLK         =       6
.equ    PORT_STR        =       PORTD           ; strobe: PD5
.equ    BIT_STR         =       5
.equ    PORT_DATA       =       PORTB           ; data bus: PB
.equ    PORT_DATA9      =       PORTD           ; 9th bit of data bus: PD2
.equ    BIT_DATA9       =       2



; register X is reserved for usage in interrupt !!!



; general purpose registers for interrupt
.def    I_TMP           =       r16
.def    I_CNT           =       r17
.def    I_REG0          =       r0
.def    I_REG1          =       r1
.def    I_REG2          =       r2

; registers for bits of 9th shift register chain in interrupt
.def    I_COL17         =       r3
.def    I_COL18         =       r4

; status register backup in interrupt
.def    I_SREG          =       r5



; general purpose registers
.def    TMP             =       r18
.def    CNT             =       r19
.def    REG0            =       r6
.def    REG1            =       r7
.def    REG2            =       r8



; current greyscale step
;  - counts 49 steps (0..48) to output 8 quadratic greyscale values
;  - for i = 0..6:
;     - in step i^2: output black/white picture, turn on pixel with value > i
.def    STEP            =       r20

; addresses of output buffers to use (see also OUT_BUF_*)
;  - flag to indicate assignment of output buffers
;    - 0         output buffer 1 for output, output buffer 2 for receiving
;    - 1 (!= 0): output buffer 2 for output, output buffer 1 for receiving
.def    OUT_PTR_FLAG    =       r21
;  - pointer to end of output buffer (used for output to LEDs)
.def    OUT_PTR_OUT_E_L =       r9
.def    OUT_PTR_OUT_E_H =       r10
;  - pointer to receive buffer (used for receiving new values)
.def    OUT_PTR_RECV_L  =       r11
.def    OUT_PTR_RECV_H  =       r12

; receive step
;  - counts bytes received so far
.def    RECV_STEP       =       r23

; receive coordinates
;  - x coordinate of next pixel
.def    RECV_X          =       r13
;  - y coordinate of next pixel
.def    RECV_Y          =       r14



.DSEG
.ORG    0x060

; output buffers with values to output (see also OUT_PTR_*)
;  - every pixel is stored as 3 bits distributed over 3 bytes
;  - the bit b ([0..2]) of the value of the pixel at x,y ([0..15],[0..7])
;    is stored in bit x/2 of byte (y+(x%2)*8)*3+b
;  - the bit b ([0..2]) of the value of the pixel at x,y ([16..17],[0..7])
;    is stored in bit y of byte x*3+b
OUT_BUF_1:      .BYTE   18*3
OUT_BUF_2:      .BYTE   18*3



.CSEG
.ORG    0x000
        rjmp    ENTRY                           ; RESET
        reti                                    ; INT0
        reti                                    ; INT1
        reti                                    ; TIMER1_CAPT
        reti                                    ; TIMER1_COMPA
        reti                                    ; TIMER1_OVF
        reti                                    ; TIMER0_OVF
        reti                                    ; USART0_RX
        reti                                    ; USART0_UDRE
        reti                                    ; USART0_TX
        reti                                    ; ANALOG_COMP
        reti                                    ; PC_INT
        reti                                    ; TIMER1_COMPB
        rjmp    INT_TIMER                       ; TIMER0_COMPA
        reti                                    ; TIMER0_COMPB
        reti                                    ; USI_START
        reti                                    ; USI_OVERFLOW
        reti                                    ; EE_READY
        reti                                    ; WDT



; code entry point
ENTRY:
; initialize stack pointer
        ldi     TMP,RAMEND
        out     SPL,TMP
; set clock prescaler to 1:1
        ldi     TMP,1<<CLKPCE
        out     CLKPR,TMP
        ldi     TMP,0
        out     CLKPR,TMP
; set ports A, B and D to output (except PD0/RXD)
        ldi     TMP,0x00
        out     PORTA,TMP
        out     PORTB,TMP
        out     PORTD,TMP
        ldi     TMP,0xFF
        out     DDRA,TMP
        ldi     TMP,0xFF
        out     DDRB,TMP
        ldi     TMP,0xFE
        out     DDRD,TMP
; disable anlaog comparator
        ldi     TMP,0x80
        out     ACSR,TMP
; enable usart
        ldi     TMP,1<<U2X                      ; double speed
        out     UCSRA,TMP
        ldi     TMP,1<<RXEN|1<<TXEN             ; receiver and transmitter enabled
        out     UCSRB,TMP
        ldi     TMP,1<<UCSZ1|1<<UCSZ0           ; asynchronous, 8N1: 8 data bits, no parity, 1 stop bit
        out     UCSRC,TMP
        ldi     TMP,0                           ; speed = Fosc / (8 * (UBRR + 1))
        out     UBRRH,TMP                       ;       = 12MHz / (8 * (12 + 1))
        ldi     TMP,12                          ;       = 115384bps (115200bps, 0.16% deviation)
        out     UBRRL,TMP
; set up timer 0 (interrupt timer)
        ldi     TMP,1<<WGM01                    ; clear timer on compare match
        out     TCCR0A,TMP
        ldi     TMP,1<<CS01|1<<CS00             ; use system clock with prescaler 1:64 (5.333us)
        out     TCCR0B,TMP
        ldi     TMP,36                          ; cycle time 36 timer steps (192us)
        out     OCR0A,TMP
        ldi     TMP,1<<OCIE0A                   ; timer 0 compare match interrupt enabled
        out     TIMSK,TMP
; set up timer 1 (timeout timer)
        ldi     TMP,0                           ; clear timer on compare match
        out     TCCR1A,TMP
        ldi     TMP,1<<CS12|1<<CS10|1<<WGM12    ; use system clock with prescaler 1:1024 (85.333us)
        out     TCCR1B,TMP
        ldi     TMP,0                           ; no output compare
        out     TCCR1C,TMP
        ldi     TMP,228                         ; timer steps 58594 (cycle time 5s)
        out     OCR1AH,TMP
        ldi     TMP,226
        out     OCR1AL,TMP
; enable watchdog (64ms)
        wdr
        ldi     TMP,1<<WDCE|1<<WDE
        out     WDTCR,TMP
        ldi     TMP,1<<WDE|1<<WDP1
        out     WDTCR,TMP
        wdr
; jump to main program
        rjmp    MAIN



; output a black/white picture
;  - turn on all pixel with value >= 1
;  - uses: X, I_COL17, I_COL18, I_REG0, I_REG1, I_REG2, I_CNT, I_TMP
;  - used stack depth (without call to this function): 0
OUT_BW_1:
        mov     XH,OUT_PTR_OUT_E_H              ; begin at end of output buffer used for output
        mov     XL,OUT_PTR_OUT_E_L
        ld      I_REG2,-X                       ; get bits 18th column (Di of values in REGi)
        ld      I_REG1,-X
        ld      I_REG0,-X
        or      I_REG2,I_REG1                   ; >= 001: D2 | D1 | D0
        or      I_REG2,I_REG0
        mov     I_COL18,I_REG2
        ld      I_REG2,-X                       ; get bits 17th column (Di of values in REGi)
        ld      I_REG1,-X
        ld      I_REG0,-X
        or      I_REG2,I_REG1                   ; >= 001: D2 | D1 | D0
        or      I_REG2,I_REG0
        mov     I_COL17,I_REG2
        ldi     I_CNT,16                        ; 16 times 8 bit left
OUT_BW_1_LOOP:
        ld      I_REG2,-X                       ; get next bits (Di of values in REGi)
        ld      I_REG1,-X
        ld      I_REG0,-X
        or      I_REG2,I_REG1                   ; >= 001: D2 | D1 | D0
        or      I_REG2,I_REG0
        out     PORT_DATA,I_REG2                ; output data
        in      I_TMP,PORT_DATA9                ; output bit for 9th shift register chain
        cbr     I_TMP,1<<BIT_DATA9
        sbrc    I_COL18,7
        sbr     I_TMP,1<<BIT_DATA9
        out     PORT_DATA9,I_TMP
        lsl     I_COL17                         ; prepare next bit for 9th shift register chain (interleaved with clock output)
        sbi     PORT_CLK,BIT_CLK                ; output clock impulse(interleaved with bit preparation)
        rol     I_COL18
        cbi     PORT_CLK,BIT_CLK
        dec     I_CNT                           ; next turn
        brne    OUT_BW_1_LOOP
        sbi     PORT_STR,BIT_STR                ; output strobe impulse
        nop
        cbi     PORT_STR,BIT_STR
        ret

; output a black/white picture
;  - turn on all pixel with value >= 2
;  - uses: X, I_COL17, I_COL18, I_REG0, I_REG1, I_REG2, I_CNT, I_TMP
;  - used stack depth (without call to this function): 0
OUT_BW_2:
        mov     XH,OUT_PTR_OUT_E_H              ; begin at end of output buffer used for output
        mov     XL,OUT_PTR_OUT_E_L
        ld      I_REG2,-X                       ; get bits 18th column (Di of values in REGi)
        ld      I_REG1,-X
        ld      I_REG0,-X
        or      I_REG2,I_REG1                   ; >= 010: D2 | D1
        mov     I_COL18,I_REG2
        ld      I_REG2,-X                       ; get bits 17th column (Di of values in REGi)
        ld      I_REG1,-X
        ld      I_REG0,-X
        or      I_REG2,I_REG1                   ; >= 010: D2 | D1
        mov     I_COL17,I_REG2
        ldi     I_CNT,16                        ; 16 times 8 bit left
OUT_BW_2_LOOP:
        ld      I_REG2,-X                       ; get next bits (Di of values in REGi)
        ld      I_REG1,-X
        ld      I_REG0,-X
        or      I_REG2,I_REG1                   ; >= 010: D2 | D1
        out     PORT_DATA,I_REG2                ; output data
        in      I_TMP,PORT_DATA9                ; output bit for 9th shift register chain
        cbr     I_TMP,1<<BIT_DATA9
        sbrc    I_COL18,7
        sbr     I_TMP,1<<BIT_DATA9
        out     PORT_DATA9,I_TMP
        lsl     I_COL17                         ; prepare next bit for 9th shift register chain (interleaved with clock output)
        sbi     PORT_CLK,BIT_CLK                ; output clock impulse(interleaved with bit preparation)
        rol     I_COL18
        cbi     PORT_CLK,BIT_CLK
        dec     I_CNT                           ; next turn
        brne    OUT_BW_2_LOOP
        sbi     PORT_STR,BIT_STR                ; output strobe impulse
        nop
        cbi     PORT_STR,BIT_STR
        ret

; output a black/white picture
;  - turn on all pixel with value >= 3
;  - uses: X, I_COL17, I_COL18, I_REG0, I_REG1, I_REG2, I_CNT, I_TMP
;  - used stack depth (without call to this function): 0
OUT_BW_3:
        mov     XH,OUT_PTR_OUT_E_H              ; begin at end of output buffer used for output
        mov     XL,OUT_PTR_OUT_E_L
        ld      I_REG2,-X                       ; get bits 18th column (Di of values in REGi)
        ld      I_REG1,-X
        ld      I_REG0,-X
        and     I_REG1,I_REG0                   ; >= 011: D2 | (D1 & D0)
        or      I_REG2,I_REG1
        mov     I_COL18,I_REG2
        ld      I_REG2,-X                       ; get bits 17th column (Di of values in REGi)
        ld      I_REG1,-X
        ld      I_REG0,-X
        and     I_REG1,I_REG0                   ; >= 011: D2 | (D1 & D0)
        or      I_REG2,I_REG1
        mov     I_COL17,I_REG2
        ldi     I_CNT,16                        ; 16 times 8 bit left
OUT_BW_3_LOOP:
        ld      I_REG2,-X                       ; get next bits (Di of values in REGi)
        ld      I_REG1,-X
        ld      I_REG0,-X
        and     I_REG1,I_REG0                   ; >= 011: D2 | (D1 & D0)
        or      I_REG2,I_REG1
        out     PORT_DATA,I_REG2                ; output data
        in      I_TMP,PORT_DATA9                ; output bit for 9th shift register chain
        cbr     I_TMP,1<<BIT_DATA9
        sbrc    I_COL18,7
        sbr     I_TMP,1<<BIT_DATA9
        out     PORT_DATA9,I_TMP
        lsl     I_COL17                         ; prepare next bit for 9th shift register chain (interleaved with clock output)
        sbi     PORT_CLK,BIT_CLK                ; output clock impulse(interleaved with bit preparation)
        rol     I_COL18
        cbi     PORT_CLK,BIT_CLK
        dec     I_CNT                           ; next turn
        brne    OUT_BW_3_LOOP
        sbi     PORT_STR,BIT_STR                ; output strobe impulse
        nop
        cbi     PORT_STR,BIT_STR
        ret

; output a black/white picture
;  - turn on all pixel with value >= 4
;  - uses: X, I_COL17, I_COL18, I_REG0, I_REG1, I_REG2, I_CNT, I_TMP
;  - used stack depth (without call to this function): 0
OUT_BW_4:
        mov     XH,OUT_PTR_OUT_E_H              ; begin at end of output buffer used for output
        mov     XL,OUT_PTR_OUT_E_L
        ld      I_REG2,-X                       ; get bits 18th column (Di of values in REGi)
        ld      I_REG1,-X
        ld      I_REG0,-X
                                                ; >= 100: D2
        mov     I_COL18,I_REG2
        ld      I_REG2,-X                       ; get bits 17th column (Di of values in REGi)
        ld      I_REG1,-X
        ld      I_REG0,-X
                                                ; >= 100: D2
        mov     I_COL17,I_REG2
        ldi     I_CNT,16                        ; 16 times 8 bit left
OUT_BW_4_LOOP:
        ld      I_REG2,-X                       ; get next bits (Di of values in REGi)
        ld      I_REG1,-X
        ld      I_REG0,-X
                                                ; >= 100: D2
        out     PORT_DATA,I_REG2                ; output data
        in      I_TMP,PORT_DATA9                ; output bit for 9th shift register chain
        cbr     I_TMP,1<<BIT_DATA9
        sbrc    I_COL18,7
        sbr     I_TMP,1<<BIT_DATA9
        out     PORT_DATA9,I_TMP
        lsl     I_COL17                         ; prepare next bit for 9th shift register chain (interleaved with clock output)
        sbi     PORT_CLK,BIT_CLK                ; output clock impulse(interleaved with bit preparation)
        rol     I_COL18
        cbi     PORT_CLK,BIT_CLK
        dec     I_CNT                           ; next turn
        brne    OUT_BW_4_LOOP
        sbi     PORT_STR,BIT_STR                ; output strobe impulse
        nop
        cbi     PORT_STR,BIT_STR
        ret

; output a black/white picture
;  - turn on all pixel with value >= 5
;  - uses: X, I_COL17, I_COL18, I_REG0, I_REG1, I_REG2, I_CNT, I_TMP
;  - used stack depth (without call to this function): 0
OUT_BW_5:
        mov     XH,OUT_PTR_OUT_E_H              ; begin at end of output buffer used for output
        mov     XL,OUT_PTR_OUT_E_L
        ld      I_REG2,-X                       ; get bits 18th column (Di of values in REGi)
        ld      I_REG1,-X
        ld      I_REG0,-X
        or      I_REG1,I_REG0                   ; >= 101: D2 & (D1 | D0)
        and     I_REG2,I_REG1
        mov     I_COL18,I_REG2
        ld      I_REG2,-X                       ; get bits 17th column (Di of values in REGi)
        ld      I_REG1,-X
        ld      I_REG0,-X
        or      I_REG1,I_REG0                   ; >= 101: D2 & (D1 | D0)
        and     I_REG2,I_REG1
        mov     I_COL17,I_REG2
        ldi     I_CNT,16                        ; 16 times 8 bit left
OUT_BW_5_LOOP:
        ld      I_REG2,-X                       ; get next bits (Di of values in REGi)
        ld      I_REG1,-X
        ld      I_REG0,-X
        or      I_REG1,I_REG0                   ; >= 101: D2 & (D1 | D0)
        and     I_REG2,I_REG1
        out     PORT_DATA,I_REG2                ; output data
        in      I_TMP,PORT_DATA9                ; output bit for 9th shift register chain
        cbr     I_TMP,1<<BIT_DATA9
        sbrc    I_COL18,7
        sbr     I_TMP,1<<BIT_DATA9
        out     PORT_DATA9,I_TMP
        lsl     I_COL17                         ; prepare next bit for 9th shift register chain (interleaved with clock output)
        sbi     PORT_CLK,BIT_CLK                ; output clock impulse(interleaved with bit preparation)
        rol     I_COL18
        cbi     PORT_CLK,BIT_CLK
        dec     I_CNT                           ; next turn
        brne    OUT_BW_5_LOOP
        sbi     PORT_STR,BIT_STR                ; output strobe impulse
        nop
        cbi     PORT_STR,BIT_STR
        ret

; output a black/white picture
;  - turn on all pixel with value >= 6
;  - uses: X, I_COL17, I_COL18, I_REG0, I_REG1, I_REG2, I_CNT, I_TMP
;  - used stack depth (without call to this function): 0
OUT_BW_6:
        mov     XH,OUT_PTR_OUT_E_H              ; begin at end of output buffer used for output
        mov     XL,OUT_PTR_OUT_E_L
        ld      I_REG2,-X                       ; get bits 18th column (Di of values in REGi)
        ld      I_REG1,-X
        ld      I_REG0,-X
        and     I_REG2,I_REG1                   ; >= 110: D2 & D1
        mov     I_COL18,I_REG2
        ld      I_REG2,-X                       ; get bits 17th column (Di of values in REGi)
        ld      I_REG1,-X
        ld      I_REG0,-X
        and     I_REG2,I_REG1                   ; >= 110: D2 & D1
        mov     I_COL17,I_REG2
        ldi     I_CNT,16                        ; 16 times 8 bit left
OUT_BW_6_LOOP:
        ld      I_REG2,-X                       ; get next bits (Di of values in REGi)
        ld      I_REG1,-X
        ld      I_REG0,-X
        and     I_REG2,I_REG1                   ; >= 110: D2 & D1
        out     PORT_DATA,I_REG2                ; output data
        in      I_TMP,PORT_DATA9                ; output bit for 9th shift register chain
        cbr     I_TMP,1<<BIT_DATA9
        sbrc    I_COL18,7
        sbr     I_TMP,1<<BIT_DATA9
        out     PORT_DATA9,I_TMP
        lsl     I_COL17                         ; prepare next bit for 9th shift register chain (interleaved with clock output)
        sbi     PORT_CLK,BIT_CLK                ; output clock impulse(interleaved with bit preparation)
        rol     I_COL18
        cbi     PORT_CLK,BIT_CLK
        dec     I_CNT                           ; next turn
        brne    OUT_BW_6_LOOP
        sbi     PORT_STR,BIT_STR                ; output strobe impulse
        nop
        cbi     PORT_STR,BIT_STR
        ret

; output a black/white picture
;  - turn on all pixel with value >= 7
;  - uses: X, I_COL17, I_COL18, I_REG0, I_REG1, I_REG2, I_CNT, I_TMP
;  - used stack depth (without call to this function): 0
OUT_BW_7:
        mov     XH,OUT_PTR_OUT_E_H              ; begin at end of output buffer used for output
        mov     XL,OUT_PTR_OUT_E_L
        ld      I_REG2,-X                       ; get bits 18th column (Di of values in REGi)
        ld      I_REG1,-X
        ld      I_REG0,-X
        and     I_REG2,I_REG1                   ; >= 111: D2 & D1 & D0
        and     I_REG2,I_REG0
        mov     I_COL18,I_REG2
        ld      I_REG2,-X                       ; get bits 17th column (Di of values in REGi)
        ld      I_REG1,-X
        ld      I_REG0,-X
        and     I_REG2,I_REG1                   ; >= 111: D2 & D1 & D0
        and     I_REG2,I_REG0
        mov     I_COL17,I_REG2
        ldi     I_CNT,16                        ; 16 times 8 bit left
OUT_BW_7_LOOP:
        ld      I_REG2,-X                       ; get next bits (Di of values in REGi)
        ld      I_REG1,-X
        ld      I_REG0,-X
        and     I_REG2,I_REG1                   ; >= 111: D2 & D1 & D0
        and     I_REG2,I_REG0
        out     PORT_DATA,I_REG2                ; output data
        in      I_TMP,PORT_DATA9                ; output bit for 9th shift register chain
        cbr     I_TMP,1<<BIT_DATA9
        sbrc    I_COL18,7
        sbr     I_TMP,1<<BIT_DATA9
        out     PORT_DATA9,I_TMP
        lsl     I_COL17                         ; prepare next bit for 9th shift register chain (interleaved with clock output)
        sbi     PORT_CLK,BIT_CLK                ; output clock impulse(interleaved with bit preparation)
        rol     I_COL18
        cbi     PORT_CLK,BIT_CLK
        dec     I_CNT                           ; next turn
        brne    OUT_BW_7_LOOP
        sbi     PORT_STR,BIT_STR                ; output strobe impulse
        nop
        cbi     PORT_STR,BIT_STR
        ret



; do step actions
;  - for i = 0..6:
;     - in step i^2: output black/white picture, turn on pixel with value > i
;  - uses: X, I_REG0, I_REG1, I_REG2, I_CNT, I_TMP
;  - used stack depth (without call to this function): 2
STEP_ACTIONS:
        cpi     STEP,10
        brsh    STEP_ACTIONS_10
        cpi     STEP,3
        brsh    STEP_ACTIONS_3
; in step 0: output black/white picture, turn on pixel with value >= 1
; in step 1: output black/white picture, turn on pixel with value >= 2
        cpi     STEP,0
        brne    STEP_ACTIONS_NOT_0
        rjmp    OUT_BW_1
STEP_ACTIONS_NOT_0:
        cpi     STEP,1
        brne    STEP_ACTIONS_NOT_1
        rjmp    OUT_BW_2
STEP_ACTIONS_NOT_1:
        ret
STEP_ACTIONS_3:
; in step 4: output black/white picture, turn on pixel with value >= 3
; in step 9: output black/white picture, turn on pixel with value >= 4
        cpi     STEP,4
        brne    STEP_ACTIONS_NOT_4
        rjmp    OUT_BW_3
STEP_ACTIONS_NOT_4:
        cpi     STEP,9
        brne    STEP_ACTIONS_NOT_9
        rjmp    OUT_BW_4
STEP_ACTIONS_NOT_9:
        ret
STEP_ACTIONS_10:
        cpi     STEP,30
        brsh    STEP_ACTIONS_30
; in step 16: output black/white picture, turn on pixel with value >= 5
; in step 25: output black/white picture, turn on pixel with value >= 6
        cpi     STEP,16
        brne    STEP_ACTIONS_NOT_16
        rjmp    OUT_BW_5
STEP_ACTIONS_NOT_16:
        cpi     STEP,25
        brne    STEP_ACTIONS_NOT_25
        rjmp    OUT_BW_6
STEP_ACTIONS_NOT_25:
        ret
STEP_ACTIONS_30:
; in step 36: output black/white picture, turn on pixel with value >= 7
        cpi     STEP,36
        brne    STEP_ACTIONS_NOT_36
        rjmp    OUT_BW_7
STEP_ACTIONS_NOT_36:
        ret



; timer interrupt
;  - every 192 us (2304 cycles)
; uses: X (reserved for interrupt), I_SREG, I_TMP, I_CNT, I_REG0, I_REG1, I_REG2, STEP
;  - used stack depth (without call to this function): 4
INT_TIMER:
        in      I_SREG,SREG
; do step actions
        rcall   STEP_ACTIONS
; increment output step number
        inc     STEP
        cpi     STEP,49
        brlo    INT_TIMER_STEP_LOWER_49
        clr     STEP
INT_TIMER_STEP_LOWER_49:
        out     SREG,I_SREG
        reti



; flip output buffers
;  - exchanges the pointer to the output buffers
;  - received output values will be output from now on
;  - restarts serial reception (i.e. discards current frame)
;  - uses: RECV_STEP, TMP
;  - used stack depth (without call to this function): 0
OUT_FLIP:
; start receiving new serial frame
;  - neccessary because output buffers will be changed
        clr     RECV_STEP
; exchange output buffers
        cpi     OUT_PTR_FLAG,0
        brne    OUT_FLIP_X
; flag 0: flag 1, output buffer 2 for output, output buffer 1 for receiving
        cli                                     ; turn off interrupts
        ldi     OUT_PTR_FLAG,1
        ldi     TMP,high(OUT_BUF_1+18*3)
        mov     OUT_PTR_OUT_E_H,TMP
        ldi     TMP,low(OUT_BUF_1+18*3)
        mov     OUT_PTR_OUT_E_L,TMP
        ldi     TMP,high(OUT_BUF_2)
        mov     OUT_PTR_RECV_H,TMP
        ldi     TMP,low(OUT_BUF_2)
        mov     OUT_PTR_RECV_L,TMP
        sei                                     ; turn on interrupts
        ret
; flag != 0: flag 0, output buffer 1 for output, output buffer 2 for receiving
OUT_FLIP_X:
        cli                                     ; turn off interrupts
        ldi     OUT_PTR_FLAG,0
        ldi     TMP,high(OUT_BUF_2+18*3)
        mov     OUT_PTR_OUT_E_H,TMP
        ldi     TMP,low(OUT_BUF_2+18*3)
        mov     OUT_PTR_OUT_E_L,TMP
        ldi     TMP,high(OUT_BUF_1)
        mov     OUT_PTR_RECV_H,TMP
        ldi     TMP,low(OUT_BUF_1)
        mov     OUT_PTR_RECV_L,TMP
        sei                                     ; turn on interrupts
        ret



; image data for LL picture
OUT_LL_DATA:
        .db     0x54, 0x98, 0xE0, 0x01, 0x00, 0x00
        .db     0x20, 0x01, 0x20, 0x01, 0x01, 0x00
        .db     0x08, 0x00, 0x09, 0x31, 0x00, 0x31
        .db     0x00, 0x01, 0x01, 0xAB, 0x67, 0x1F
        .db     0xAA, 0xCC, 0xF0, 0x00, 0x00, 0x00
        .db     0x08, 0x00, 0x08, 0x00, 0x00, 0x00
        .db     0x20, 0x00, 0x20, 0x18, 0x00, 0x18
        .db     0x00, 0x00, 0x00, 0x55, 0x33, 0x0F
        .db     0x01, 0x01, 0x01, 0x55, 0x33, 0x0F



; show LL picture
;  - uses: Y, Z, CNT, TMP
;  - used stack depth (without call to this function): 0
OUT_LL:
        ldi     ZH,high(2*OUT_LL_DATA)          ; load base address of image ("2*" to convert word address to byte address)
        ldi     ZL,low(2*OUT_LL_DATA)
        mov     YH,OUT_PTR_RECV_H               ; get address of output buffer to put new data in
        mov     YL,OUT_PTR_RECV_L
        ldi     CNT,18*3
OUT_LL_LOOP:
        lpm     TMP,Z+                          ; read byte from program memory
        st      Y+,TMP                          ; write byte to output buffer
        dec     CNT
        brne    OUT_LL_LOOP
        rjmp    OUT_FLIP                        ; show new picture



; serial reception
; uses: Z, CNT, TMP, REG0, REG1, RECV_STEP
;  - used stack depth (without call to this function): 0
RECV:
; check for reception
        sbis    UCSRA,RXC
        rjmp    RECV_END                        ; nothing received
; check for errors
        sbic    UCSRA,FE
        rjmp    RECV_ERR                        ; frame error
        sbic    UCSRA,DOR
        rjmp    RECV_ERR                        ; data overrun
; check if behind message
        cpi     RECV_STEP,12+18*8
        brlo    RECV_STEP_OK
        clr     RECV_STEP                       ; behind message -> start receiving a new message
RECV_STEP_OK:        
; check if in header or in data part
        cpi     RECV_STEP,12
        brlo    RECV_HEADER
        rjmp    RECV_DATA

; receive error
RECV_ERR:
        in      TMP,UDR                         ; read data from USART to clear error
        clr     RECV_STEP                       ; start receiving a new message
RECV_END:
        ret

; expected header data (MCUF header 18x8-1/256)
RECV_HEADER_DATA:
        .db     0x23, 0x54, 0x26, 0x66          ; magic (first byte is also contained in code)
        .db     0x00, 0x08                      ; height: 8
        .db     0x00, 0x12                      ; width: 18
        .db     0x00, 0x01                      ; channels: 1
        .db     0x00, 0xFF                      ; maxval: 255

; received a header byte
RECV_HEADER:
        in      REG0,UDR                        ; fetch data byte from USART
        ldi     ZH,high(2*RECV_HEADER_DATA)     ; load base address of header data ("2*" to convert word address to byte address)
        ldi     ZL,low(2*RECV_HEADER_DATA)
        clr     TMP                             ; add offset
        add     ZL,RECV_STEP
        adc     ZH,TMP
        lpm     REG1,Z                          ; fetch expected data byte from program memory
        cp      REG0,REG1                       ; compare with received data byte
        breq    RECV_HEADER_OK
        clr     RECV_STEP                       ; mismatch -> start receiving new message
        ldi     TMP,0x23                        ; check if first byte of message (first byte of magic)
        cp      REG0,TMP
        breq    RECV_HEADER_OK                  ; first byte of message/magic -> accept it
        ret
RECV_HEADER_OK:
        inc     RECV_STEP                       ; match -> next byte
        clr     RECV_X                          ; next pixel to receive is topleft pixel
        clr     RECV_Y                          ;  - only needed for last header byte, but faster and more easy this way
        ret

; received a data byte
RECV_DATA:
        in      REG0,UDR                        ; fetch data byte from USART
        mov     ZH,OUT_PTR_RECV_H               ; get base address of output buffer to put received data in
        mov     ZL,OUT_PTR_RECV_L
        sbrs    RECV_X,4                        ; skip if received byte is at x coordinate 16..17
        rjmp    RECV_DATA_0_15                  ; received byte is at x coordinate 0..15
        mov     REG1,RECV_X                     ; byte_3 (REG1) := x
        mov     REG2,RECV_Y                     ; bit := y
        rjmp    RECV_DATA_CONT
RECV_DATA_0_15:
        mov     REG1,RECV_Y                     ; byte_3 (REG1) := y + (x % 2) * 8
        ldi     TMP,8
        sbrc    RECV_X,0
        add     REG1,TMP
        mov     REG2,RECV_X                     ; bit (REG2) := x / 2
        lsr     REG2
RECV_DATA_CONT:
        mov     TMP,REG1                        ; byte (REG1) := byte_3 (REG1) * 3
        add     REG1,TMP
        add     REG1,TMP
        clr     TMP                             ; get destination address (dest_addr := base_addr + byte)
        add     ZL,REG1
        adc     ZH,TMP
        ldi     TMP,1                           ; get bit mask into REG2
        sbrc    REG2,0
        lsl     TMP
        sbrc    REG2,1
        lsl     TMP
        sbrc    REG2,1
        lsl     TMP
        sbrc    REG2,2
        swap    TMP
        mov     REG2,TMP
        mov     REG1,REG2                       ; get negated bit mask into REG1
        com     REG1
        ld      TMP,Z                           ; store D0 of pixel value (D5 of received data byte)
        and     TMP,REG1
        sbrc    REG0,5
        or      TMP,REG2
        st      Z+,TMP
        ld      TMP,Z                           ; store D1 of pixel value (D6 of received data byte)
        and     TMP,REG1
        sbrc    REG0,6
        or      TMP,REG2
        st      Z+,TMP
        ld      TMP,Z                           ; store D2 of pixel value (D7 of received data byte)
        and     TMP,REG1
        sbrc    REG0,7
        or      TMP,REG2
        st      Z+,TMP
        inc     RECV_STEP                       ; next byte
        inc     RECV_X                          ; next pixel
        ldi     TMP,18
        cp      RECV_X,TMP
        brlo    RECV_NEXT_END
        clr     RECV_X                          ; next line
        inc     RECV_Y
RECV_NEXT_END:
        cpi     RECV_STEP,12+18*8               ; end of frame - output picture
        brsh    RECV_DATA_OUTPUT
        ret

; output picture
RECV_DATA_OUTPUT:
; reset timeout timer
        cli                                     ; turn off interrupts
        clr     TMP                             ; reset timer 1 (timeout timer)
        out     TCNT1H,TMP
        out     TCNT1L,TMP
        in      TMP,OCF1A                       ; clear timeout flag (timer 1 compare match)
        sbr     TMP,1<<OCF1A
        out     TIFR,TMP
        sei                                     ; turn on interrupts
; show received data
        rjmp    OUT_FLIP                        ; show received data



; main program - initialization
;  - used stack depth: 2
MAIN:
        wdr                                     ; watchdog reset
        clr     STEP                            ; output step counter to first output step
        clr     OUT_PTR_FLAG                    ; initialize output buffer pointers
        rcall   OUT_FLIP
        rcall   OUT_LL                          ; show LL picture
        clr     RECV_STEP                       ; start receiving with a new message
        clr     TMP                             ; reset timer 1 (timeout timer)
        out     TCNT1H,TMP
        out     TCNT1L,TMP
        in      TMP,OCF1A                       ;  - clear timeout flag (timer 1 compare match)
        sbr     TMP,1<<OCF1A
        out     TIFR,TMP
        sei                                     ; turn on interrupts



; main program - main loop
MAIN_LOOP:

; watchdog reset
        wdr

; receive serial data
        rcall   RECV

; timeout -> LL picture
        in      TMP,TIFR                        ; skip if timeout flag (timer 1 compare match) is set
        sbrs    TMP,OCF1A
        rjmp    MAIN_TIMEOUT_END
        rcall   OUT_LL                          ; show LL picture
        cli                                     ; turn off interrupts
        in      TMP,OCF1A                       ; clear timeout flag (timer 1 compare match)
        sbr     TMP,1<<OCF1A
        out     TIFR,TMP
        sei                                     ; turn on interrupts
MAIN_TIMEOUT_END:

; bottom of main loop
        rjmp     MAIN_LOOP

